# -*- coding:utf-8 -*-
from star_builder import Type, validators
import uuid
import datetime



class GroupFillingRate(Type):
    """队列填充率"""
    group_id = validators.String()
    enum_list = validators.Array(items=validators.String())
    total_value_count = validators.Integer(minimum=0)
    valid_value_count = validators.Integer(minimum=0)
    invalid_value_count = validators.Integer(minimum=0)
    value = validators.Number()


class FillingRate(Type):
    """填充率"""
    update_time = validators.DateTime(allow_null=True)
    total_value_count = validators.Integer(minimum=0, default=0)
    valid_value_count = validators.Integer(minimum=0, default=0)
    invalid_value_count = validators.Integer(minimum=0, default=0)
    value = validators.Number(default=0)
    enum_list = validators.Array(items=validators.String(), default=list)
    group = validators.Array(items=GroupFillingRate, default=list)

    @classmethod
    def has_default(cls):
        return True


class Variable(Type):
    """统计变量"""
    _id = validators.String(format="UUID", default=uuid.uuid4)
    desease_id = validators.String(default="")
    project_id = validators.String(default="")
    name = validators.String(default="")
    parent_project_key = validators.String(default="")  # 父级project_key
    project_key = validators.String(default="")  # project_key
    row_index = validators.Integer(minimum=0, allow_null=True)  # 组自增题行index
    follow_group_id = validators.String(default="")  # 随访队列id
    follow_stage_id = validators.String(default="")  # 随访阶段id
    follow_round = validators.Integer(minimum=0, allow_null=True)  # 随访点
    follow_parent_ques_id = validators.String(default="")  # 随访父级题目id
    follow_ques_id = validators.String(default="")  # 随访题目id
    is_group_variable = validators.Boolean(default=False)  # 是否是组别变量
    order = validators.Integer(minimum=0, default=0)  # 排序
    create_time = validators.DateTime(default=datetime.datetime.now)
    filling_rate = FillingRate  # 填充率

a = Variable({})
a.format()
import pdb;pdb.set_trace()
print(a)
b = a.to_json()
c = Variable(b)
c.format()
print(c)
d = c.to_json()
e = Variable(d)
e.format()
print(e)
# import json
# b = json.dumps({'phase_name': 'basic_name'})
# print(json.loads(b.))




import os
import sys
import uvloop
import signal
import asyncio
import logging
import traceback

from toolkit import cache_property
from argparse import ArgumentParser

from apistar import Route
from apistar.http import PathParams, Response
from apistar.server.injector import ASyncInjector
from apistar.server.validation import VALIDATION_COMPONENTS
from apistar.server.asgi import ASGI_COMPONENTS, ASGIReceive, \
    ASGIScope, ASGISend

from . import Solo
from ..bases.components import SettingsComponent
from ..helper import find_children, ArgparseHelper, load_packages


class MySelf(object):
    def __getattr__(self, item):
        return self

    def __getitem__(self, item):
        if item == 'type':
            return "http.request"
        return self

    def __call__(self, *args, **kwargs):
        return self

    def __bool__(self):
        return False

    def __str__(self):
        return ""

    def __radd__(self, obj):
        return obj + type(obj)(self)

    def __iter__(self):
        return iter([])

    def __await__(self):
        loop = asyncio.get_event_loop()
        future = loop.create_future()
        future.set_result(self)
        return future.__await__()

    __repr__ = __str__


class SoloManager(object):
    """
    独立任务程序管理器
    """
    alive = True

    def __init__(self, app_name, current_dir="."):
        os.chdir(current_dir)
        sys.path.insert(0, current_dir)
        sys.modules.pop(app_name, None)
        load_packages(".")
        self.solos = {solo.__name__.lower(): solo
                      for solo in find_children(Solo, False)}
        self.args = self.parse_args()
        SettingsComponent.register_path(self.args.settings)
        initial_components = {
            'scope': ASGIScope,
            'receive': ASGIReceive,
            'send': ASGISend,
            'exc': Exception,
            'app': SoloManager,
            'path_params': PathParams,
            'route': Route,
            'response': Response,
        }
        self.state = {
                'scope': MySelf(),
                'receive': MySelf(),
                'send': MySelf(),
                'exc': None,
                'app': self,
                'path_params': MySelf(),
                'route': MySelf()
            }
        self.injector = ASyncInjector(
            list(ASGI_COMPONENTS + VALIDATION_COMPONENTS) + find_children(),
            initial_components)
        self.solo = self.solos[self.args.solo](**vars(self.args))
        self.task = None

    @cache_property
    def logger(self):
        """
        可以选择覆盖这个属性
        :return:
        """
        return logging.getLogger("solo")

    def start(self):
        asyncio.get_event_loop().close()
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGQUIT, self.handle_exit,
                                signal.SIGQUIT, None)
        loop.add_signal_handler(signal.SIGTERM, self.handle_exit,
                                signal.SIGTERM, None)
        loop.add_signal_handler(signal.SIGINT, self.handle_exit,
                                signal.SIGINT, None)
        loop.add_signal_handler(signal.SIGABRT, self.handle_exit,
                                signal.SIGABRT, None)

        self.task = loop.create_task(
            self.injector.run_async(
                [self.solo.setup, self.solo.run], dict(self.state)))
        loop.create_task(self.tick(loop))

        self.logger.info(f'Starting worker [{os.getpid()}]')

        loop.run_forever()

    async def tick(self, loop):
        while self.alive:
            if self.task.done():
                break
            for task in self.solo.tasks[:]:
                try:
                    if task.done():
                        rs = task.result()
                        if rs is not None:
                            self.logger.info(rs)
                        self.solo.tasks.remove(task)
                except Exception as e:
                    self.logger.error(f"Error in task: {task} error: {e}")
                    self.solo.tasks.remove(task)

            await asyncio.sleep(1)
        try:
            if self.task.done():
                rs = self.task.result()
                if rs is not None:
                    self.logger.info(rs)
        except Exception as e:
            self.logger.error(
                f"Error in main coroutine, error: {traceback.format_exc()}")

        await self.injector.run_async(
            [self.solo.teardown], dict(self.state))
        self.logger.warning(f"Stopping [{os.getpid()}]")
        loop.stop()

    def handle_exit(self, sig, frame):
        self.alive = False
        self.logger.warning(
            "Received signal {}. Shutting down.".format(sig.name))

    def parse_args(self):
        base_parser = ArgumentParser(
            description=self.__class__.__doc__, add_help=False)
        base_parser.add_argument(
            "--settings", help="配置模块路径.", default="settings")

        parser = ArgumentParser(description="独立任务程序构建工具", add_help=False)
        parser.add_argument(
            '-h', '--help', action=ArgparseHelper, help='显示帮助信息并退出.')
        sub_parsers = parser.add_subparsers(dest="solo", help="创建独立任务服务类型.")

        for name, solo in self.solos.items():
            sub_parser = sub_parsers.add_parser(
                name.lower(), parents=[base_parser], help=solo.__doc__)
            solo.enrich_parser(sub_parser)
        # 当不提供参数时，len(sys.argv) == 1, 不会打印帮助，需要手动打印
        # add_subparsers无法指定required=True, 这是argparse的bug。
        if len(sys.argv) < 2:
            parser.print_help()
            exit(1)
        return parser.parse_args()


import os
import sys
import asyncio
import logging
import traceback

from apistar import ASyncApp, App, exceptions
from apistar.server.asgi import ASGIScope, ASGISend
from apistar.http import Response, JSONResponse
from apistar.server.components import ReturnValue

from .bases.controller import Controller
from .bases.components import SettingsComponent
from .bases.hooks import ErrorHook, AccessLogHook, SessionHook, Hook
from .helper import load_packages, routing, print_routing, TypeEncoder, \
    find_children, enhance_response

__all__ = ["Application"]
enhance_response(Response)
JSONResponse.options["default"] = TypeEncoder().default


class FixedAsyncApp(ASyncApp):

    def exception_handler(self, exc: Exception) -> Response:
        if isinstance(exc, exceptions.HTTPException):
            payload = {
                "type": "http",
                "code": exc.status_code,
                "errcode": exc.status_code,
                "message": exc.detail,
            }
            if self.debug:
                payload["detail"] = "".join(traceback.format_exc())
            return JSONResponse(payload, exc.status_code, exc.get_headers())
        raise exc

    def error_handler(self, return_value: ReturnValue) -> Response:
        if isinstance(return_value, Response):
            return return_value
        return super().error_handler()

    async def read(self, response):
        coroutine = response.content.read(1024000)
        if asyncio.iscoroutine(coroutine):
            return await coroutine
        else:
            return coroutine

    async def finalize_asgi(self,
                            response: Response,
                            send: ASGISend,
                            scope: ASGIScope):
        if response.exc_info is not None:
            if self.debug or scope.get('raise_exceptions', False):
                exc_info = response.exc_info
                raise exc_info[0].with_traceback(exc_info[1], exc_info[2])

        await send({
            'type': 'http.response.start',
            'status': response.status_code,
            'headers': [
                [key.encode(), value.encode()]
                for key, value in response.headers
            ]
        })
        if hasattr(response.content, "read"):
            body = await self.read(response)
            while body:
                await send({
                    'type': 'http.response.body',
                    'body': body,
                    "more_body": True,
                })
                body = await self.read(response)
        else:
            body = response.content
        await send({
            'type': 'http.response.body',
            'body': body
        })


class FixedApp(App):

    def exception_handler(self, exc: Exception) -> Response:
        if isinstance(exc, exceptions.HTTPException):
            payload = {
                "type": "http",
                "code": exc.status_code,
                "errcode": exc.status_code,
                "message": exc.detail,
            }
            if self.debug:
                payload["detail"] = "".join(traceback.format_exc())
            return JSONResponse(payload, exc.status_code, exc.get_headers())
        raise exc

    def error_handler(self, return_value: ReturnValue) -> Response:
        if isinstance(return_value, Response):
            return return_value
        return super().error_handler()


def application(app_name, template_dir=None,
                static_dir=None,
                packages=None,
                schema_url='/schema/',
                docs_url='/docs/',
                static_url='/static/',
                settings_path="settings",
                debug=True,
                async=True,
                current_dir="."):
    """
       参数指定选择使用异步app还是同步app
       可以动态发现当前项目根目录下所有controller中的handler
    """
    logger = logging.getLogger(app_name)
    os.chdir(current_dir)
    sys.path.insert(0, current_dir)
    sys.modules.pop(app_name, None)
    load_packages(".")
    include = routing(Controller)
    SettingsComponent.register_path(settings_path)

    if include:
        routes = [include]
        print_routing(routes, write=logger.debug)
    else:
        logger.info("Noting to route. ")
        routes = []
    custom_hooks = sorted(find_children(Hook), key=lambda x: x.order)
    hooks = [AccessLogHook(), SessionHook(), ErrorHook()] + custom_hooks
    app = (FixedAsyncApp if async else FixedApp)(
        routes,
        template_dir=template_dir,
        static_dir=static_dir,
        packages=packages,
        schema_url=schema_url,
        docs_url=docs_url,
        static_url=static_url,
        components=find_children(),
        event_hooks=hooks)

    app.debug = debug
    return app


Application = application


def show_routes():
    sys.path.insert(0, ".")
    formatter = "{:<40} {:<7} {:<40} {:<}"

    def show_format(method, parttern, name, ca_name):
        return formatter.format(name, method, parttern, ca_name)

    load_packages(".")
    include = routing(Controller)
    if include:
        print(formatter.format(
            "Name", "Method", "URI Pattern", "Controller#Action"))
        print_routing([include], format=show_format)
    else:
        print("Noting to route. ")



from star_builder import Type, validators


class Subject(Type):
    id = validators.String(allow_null=True)


class Student(Type):
    name = validators.String()
    subject = validators.Proxy(Subject, allow_null=True)


class Teacher(Type):
    name = validators.String()
    subject = validators.Proxy(Subject, default=dict)


a = Student({"name": "tom", "subject": {"id": "10"}})

print(a)

a.format()
print(a)

b = Student(name="tom")
print(b)
b.format()
print(b)

c = Teacher(name="john")
print(c)
c.format()
print(c)

d = Teacher(b.to_dict())
print(d)
d.format()
print(d)


import requests

url = "http://127.0.0.1:8000/test_download"

resp = requests.get(url, params={"filename": "/Users/mashichao/Downloads/zgtj.zip"}, stream=True)

with open("a.zip", "wb") as f:
    for chunck in resp.iter_content(1024000):
        f.write(chunck)


import sys
import typing
import logging
import traceback

from apistar import App, http
from flask.sessions import SecureCookieSessionInterface

from .session import Session
from ..helper import HookReturn
from .components import DummyFlaskApp


class SessionHook(object):

    def __init__(self):
        self.session_interface = SecureCookieSessionInterface()

    def on_response(self,
                    app: DummyFlaskApp,
                    resp: http.Response,
                    session: Session):
        if session is not None:
            self.session_interface.save_session(app, session, resp)


class ErrorHook(object):
    """
    处理异常
    """
    errors = {999: "Unknown error"}

    def on_error(self, error: Exception, app: App) -> http.Response:
        """
        Handle error
        """
        code = 999
        message = None
        if error.args:
            if not isinstance(error.args[0], (tuple, list)) \
                    or len(error.args[0]) < 2:
                if isinstance(error.args[0], int):
                    code = error.args[0]
                else:
                    message = error.args[0]
            else:
                code, message = error.args[0][:2]

        code = int(code)
        # apistar不支持在on_request时打断后续执行直接返回response
        # 所以在只能通过raise异常来通过异常参数传递响应。
        if isinstance(message, http.Response):
            return message

        if message is None:
            message = self.errors.get(code, "Not configured error")

        payload = {
            "type": "normal",
            "code": code,
            "errcode": code,
            "message": message,
        }
        if app.debug:
            payload["detail"] = "".join(traceback.format_exc())
        traceback.print_exc()
        return http.JSONResponse(payload)

    @classmethod
    def register(cls,
                 errors: typing.Union[typing.List[typing.Tuple[int, str]],
                                      typing.Mapping[int, str]]):
        cls.errors.update(errors)


class AccessLogHook(object):
    fmt = '{host} - - [{asctime}] {method} {path} {protocol}' \
          ' {status} {content_length} {agent}'

    def __init__(self):
        self.logger = logging.getLogger("access")
        self.logger.propagate = False
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(10)
        handler.setFormatter(logging.Formatter(self.fmt, style="{"))
        self.logger.addHandler(handler)

    def on_response(self,
                    host: http.Host,
                    path: http.Path,
                    protocol: http.Scheme,
                    method: http.Method,
                    resp: http.Response,
                    user_agent: http.Header):
        self.log(host, path, protocol, method, resp.status_code,
                 resp.headers["Content-Length"], user_agent)

    def on_error(self,
                 host: http.Host,
                 path: http.Path,
                 protocol: http.Scheme,
                 method: http.Method,
                 resp: http.Response,
                 user_agent: http.Header) -> http.Response:
        self.log(host, path, protocol, method, resp.status_code,
                 resp.headers["Content-Length"], user_agent)
        return resp

    def log(self, host, path, protocol, method, status, content_length, agent):
        self.logger.info("", extra={"host": host,
                                    "path": path,
                                    "protocol": protocol,
                                    "method": method,
                                    "status": status,
                                    "content_length": content_length,
                                    "agent": agent})


class Hook(object):
    """
    Hook基类，继承自此基类的hook可以自动发现。
    """
    order = 1


def Return(return_value):
    if isinstance(return_value, str):
        return_value = http.HTMLResponse(return_value)
    elif not isinstance(return_value, http.Response):
        return_value = http.JSONResponse(return_value)
    raise HookReturn(return_value)


from star_builder import Type, validators


class A(Type):
    name = validators.String()

class Union(Type):
    field = (validators.Array(validators.String(default="")) | validators.String(
        default="")) << {"allow_null": True}
    print(field)
    print(field.allow_null)
    # name = validators.Proxy(A, allow_null=True)
a = Union()
a.format()
print(a)




import typing
import inspect

from inspect import Parameter
from datetime import timedelta
from collections import namedtuple
from urllib.parse import unquote

from toolkit.singleton import Singleton
from toolkit.frozen import FrozenSettings
from toolkit.settings import SettingsLoader

from werkzeug.datastructures import ImmutableMultiDict
from werkzeug.formparser import FormDataParser
from werkzeug.http import parse_options_header
from flask.sessions import SecureCookieSessionInterface

from apistar.server.asgi import ASGIReceive
from apistar.conneg import negotiate_content_type
from apistar import Route, exceptions, http, Component as _Component, codecs

from .session import Session
from .controller import Controller


class Component(_Component, metaclass=Singleton):

    def identity(self, parameter: inspect.Parameter):
        """
        Each component needs a unique identifier string that we use for lookups
        from the `state` dictionary when we run the dependency injection.
        """
        parameter_name = parameter.name.lower()
        annotation_name = str(parameter.annotation)

        # If `resolve_parameter` includes `Parameter` then we use an identifier
        # that is additionally parameterized by the parameter name.
        args = inspect.signature(self.resolve).parameters.values()
        if inspect.Parameter in [arg.annotation for arg in args]:
            return annotation_name + ':' + parameter_name

        # Standard case is to use the class name, lowercased.
        return annotation_name

    def resolve(self, *args, **kwargs) -> object:
        raise NotImplementedError()

    def can_handle_parameter(self, parameter: inspect.Parameter):
        """重写这个方法是为了增加typing.Union类型的判定"""
        return_annotation = inspect.signature(self.resolve).return_annotation
        if return_annotation is inspect.Signature.empty:
            msg = (
                'Component "%s" must include a return annotation on the '
                '`resolve()` method, or override `can_handle_parameter`'
            )
            raise exceptions.ConfigurationError(msg % self.__class__.__name__)
        return type(return_annotation) == typing._Union and \
               parameter.annotation in return_annotation.__args__ or \
               parameter.annotation is return_annotation


class ServiceComponent(Component):
    """
    注入Service
    """
    def resolve(self, route: Route) -> Controller:
        return route.controller


class SettingsComponent(Component):
    """
    注入Settings
    """
    settings_path = None

    def __init__(self):
        self.settings = SettingsLoader().load(self.settings_path or "settings")

    def resolve(self) -> FrozenSettings:
        return self.settings

    @classmethod
    def register_path(cls, settings_path):
        cls.settings_path = settings_path


Cookie = typing.NewType('Cookie', str)


class CookiesComponent(Component):
    def resolve(self, cookie: http.Header) -> typing.Dict[str, Cookie]:
        cookies = dict()
        if cookie:
            for c in cookie.split(";"):
                key, val = c.strip().split("=", 1)
                cookies[key] = Cookie(val)
        return cookies


class CookieComponent(Component):

    def resolve(self,
                parameter: Parameter,
                cookies: typing.Dict[str, Cookie]) -> Cookie:
        return cookies.get(parameter.name.replace('_', '-'))


DummyFlaskApp = namedtuple(
    "DummyFlaskApp",
    "session_cookie_name,secret_key,permanent_session_lifetime, config")


class DummyFlaskAppComponent(Component):

    def __init__(self):
        self.default_config = {
            'SESSION_COOKIE_NAME': 'session',
            'SESSION_COOKIE_DOMAIN': None,
            'SESSION_COOKIE_PATH': '/',
            'SESSION_COOKIE_HTTPONLY': True,
            'SESSION_COOKIE_SECURE': False,
            'SESSION_COOKIE_SAMESITE': None,
            'SESSION_REFRESH_EACH_REQUEST': True,
            'SERVER_NAME': None,
        }

    def resolve(self, settings: FrozenSettings,
                host: http.Header,
                ip_host: http.Host) -> DummyFlaskApp:
                self.default_config["SERVER_NAME"] = host or ip_host
                return DummyFlaskApp(
                    config=settings.get("SESSION_CONFIG", self.default_config),
                    secret_key=settings.get("SECRET_KEY"),
                    permanent_session_lifetime=timedelta(
                        days=settings.get("PERMANENT_SESSION_LIFETIME", 31)),
                    session_cookie_name=settings.get(
                        "SESSION_COOKIE_NAME", "session"))


class SessionComponent(Component):

    def __init__(self):
        self.dummy_request = namedtuple("Request", "cookies")
        self.session_interface = SecureCookieSessionInterface()

    def resolve(self,
                app: DummyFlaskApp,
                cookies: typing.Dict[str, Cookie]) -> Session:
            request = self.dummy_request(cookies=cookies)
            return self.session_interface.open_session(app, request)


class File(object):

    def __init__(self, stream, receive, boundary, name, filename):
        self.receive = receive
        self.filename = filename
        self.name = name
        self.stream = stream
        self.tmpboundary = b"\r\n--" + boundary
        self.last = b"\r\n--" + boundary + b"--\r\n"

    def __aiter__(self):
        return self.iter_content()

    async def iter_content(self):
        body = self.stream.body
        while True:
            index = body.find(self.tmpboundary)
            if index != -1:
                read, self.stream.body = body[:index], body[index:]
                yield read
                break
            else:
                if self.stream.closed:
                    raise RuntimeError("Uncomplete content!")
                read, body = body[:-len(self.tmpboundary)], body[-len(self.tmpboundary):]
                yield read
                message = await self.get_message(self.receive)
                body += message.get('body', b'')
                if not message.get('more_body', False):
                    self.stream.closed = True

    async def read(self, size=10240):
        assert size > 0, (999, "Read size must > 0")
        _size = size
        body = self.stream.body
        while True:
            if len(body) < size + len(self.tmpboundary):
                if not self.stream.closed:
                    message = await self.get_message(self.receive)
                    body += message.get('body', b'')
                    if not message.get('more_body', False):
                        self.stream.closed = True
                    continue
                else:
                    _size = len(body) - len(self.last)
            break
        index = body.find(self.tmpboundary)
        if index != -1:
            _size = index

        read, self.stream.body = body[:_size], body[_size:]
        return read

    @staticmethod
    async def get_message(receive):
        message = await receive()
        if not message['type'] == 'http.request':
            error = "'Unexpected ASGI message type '%s'."
            raise Exception(error % message['type'])

        return message

    @classmethod
    async def from_boundary(cls, stream, receive, boundary):
        tmp_boundary = b"--" + boundary
        end_boundary = b"--" + boundary + b"--"
        while not stream.closed:
            message = await cls.get_message(receive)
            stream.body += message.get('body', b'')
            if b"\r\n\r\n" in stream.body and tmp_boundary in stream.body or \
                    not message.get('more_body', False):
                break

            if not message.get('more_body', False):
                stream.closed = True
            else:
                stream.closed = False

        stream.body, name, filename = cls.parse_headers(
            stream.body, tmp_boundary, end_boundary)
        return cls(stream, receive, boundary, name, filename)

    @staticmethod
    def parse_headers(body, tmp_boundary, end_boundary):
        index = body.find(tmp_boundary)
        if index == body.find(end_boundary):
            raise StopAsyncIteration

        body = body[index + len(tmp_boundary):]
        header_str = body[:body.find(b"\r\n\r\n")]
        body = body[body.find(b"\r\n\r\n") + 4:]
        filename = ""
        name = ''
        for header in header_str.split(b"\r\n"):
            if header.startswith(b"Content-Disposition"):
                for d in header.split(b";"):
                    d = d.strip()
                    if b"=" in d:
                        k, v = d.split(b'=')
                        if k == b'name':
                            name = v.strip(b"\"")
                        elif k == b'filename':
                            filename = v.strip(b"\"")
                        elif k == b"filename*":
                            # 用来处理带编码的文件名，返回unicode
                            enc, lang, fn = v.split(b"'")
                            filename = unquote(fn).decode(enc)
        return body, name, filename


class FileStream(object):

    def __init__(self, receive, boundary):
        self.receive = receive
        self.boundary = boundary
        self.body = b""
        self.closed = False

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await File.from_boundary(self, self.receive, self.boundary)


class FileStreamComponent(Component):
    media_type = 'multipart/form-data'

    async def decode(self, receive, headers):
        try:
            mime_type, mime_options = parse_options_header(
                headers['content-type'])
        except KeyError:
            mime_type, mime_options = '', {}

        boundary = mime_options.get('boundary', "").encode()
        if boundary is None:
            raise ValueError('Missing boundary')

        return FileStream(receive, boundary)

    async def resolve(self,
                      receive: ASGIReceive,
                      headers: http.Headers,
                      content_type: http.Header) -> FileStream:
        try:
            negotiate_content_type([self], content_type)
        except exceptions.NoCodecAvailable:
            raise exceptions.UnsupportedMediaType()

        return await self.decode(receive, headers)



# -*- coding:utf-8 -*-
import json
import time
import datetime
from star_builder.types import Type, validators, TypeEncoder


class Book(Type):
    # 使用正则限制名称格式
    name = validators.String(pattern=r"<<.*?>>")
    # 枚举类型
    publisher = validators.String(enum=["新华出版社", "人民教育出版社", "人民邮电出版社"])
    # 作者最大长度不能超20
    author = validators.String(max_length=20)
    # 创建日期类型%Y-%m-%d 目前只支持这一种，可能自定义类型
    publish_date = validators.Date(default=datetime.datetime.now)
    # 不能小于100页
    page_num = validators.Integer(minimum=100)
    # float类型支持
    price = validators.Number()
    # 整除支持, 要2本2本买
    sale_per_count = validators.Integer(multiple_of=2)
    # 支持%Y-%m-%d %H:%M:%S
    created_at = validators.DateTime()
    # 是否在售
    on_sale = validators.Boolean()
    # 当allow_null=True时，若其值为None, 如果指定了default，会使用default
    description = validators.String(allow_null=True, default="这是一段描述")

# 原始json对象
book = {
        "page_num": 628,
        "price": 139.00,
        "sale_per_count": 2,
        "created_at": time.strftime("%Y-%m-%d %H:%M:%S"),
        "on_sale": True,
        "description": None
        }
# 创建Book对象， 当指定Book(book, force_format=True)时，会强制校验数据合法性
b = Book(book)
# 赋值时会校验数据合法性
b.name = "13444" # 这条数据会报错

# 下面的数据不会报错
b.name = "<<流畅的python>>"
b.publisher = "人民邮电出版社"
b.author = "Luciano Ramalho"
b.publish_date = "2017-5-15"

# 直接dumps,需要指定cls=TypeEncoder, 同时会校验数据合法性
print(json.dumps(b, cls=TypeEncoder, indent=2))
# to_json返回json字典，同时会校验数据合法性，当to_json(force_format=False)时，数据合法性不会校验
print(b.to_json())


# model可以从对象中获取
class A:
    def __init__(self):
        self.name = "<<流畅的python>>"
        self.publisher = "人民邮电出版社"
        self.author = "Luciano Ramalho"
a = A()
t = Book(a)
print(json.dumps(t, cls=TypeEncoder))


class Name(Type):
    first_name = validators.String()
    last_name = validators.String()
    # 可以引用自身
    full_name = validators.Ref("Name", default="全名")

    # 当Name作为其它对象的属性时，可以允许是Null
    @classmethod
    def allow_null(cls):
        return True


class Tape(Type):
    hours = validators.Integer()
    name = validators.String()


class Pad(Type):
    wight = validators.Number()
    brand = validators.String()


class People(Type):
    name = Name
    # exclusive_maximum指定包含边界，即len(id) <= 32
    id = validators.Integer(maximum=32, exclusive_maximum=True)
    # 当items是People时，Array中的对象全是People
    lovers = validators.Array(items=validators.Ref("People"))
    # 最多两项好爱好，第一项只能是唱片，第二项只能是Pad，不接受大于items长度的项
    hobbies = validators.Array(items=[Tape, Pad], additional_items=False)


people = {"name": {"first_name": "Tom",
                     "last_name": "Cat",
                     "full_name": None},
            "id": 13,
            "lovers": [{"name": {
                "first_name": "Bill",
                "last_name": "Gate",
                "full_name": {"first_name": "bill", "last_name": "gate"}
            },
            "id": 1, "lovers": [], "hobbies": [{"hours": 2, "name": "my heart will go on "}]}],
            "hobbies": [{"hours": 3, "name": "audio"}, {"wight": 3, "brand": "apple"}]}

p = People(people, force_format=True)
print(p.lovers[0].name.full_name)


class Ipad(Pad):
    # 覆盖父类属性
    brand = validators.String(enum=["apple"])
    # 创建一个union对象，price可以是Number类型或String类型
    price = validators.Union([validators.Number(), validators.String()])


class Box(Type):
    ipads = validators.Array(items=Ipad)


box = {"ipads": [{"wight": 10.1,
        "brand": "apple",
        "price": 10.0}, {"wight": 10.1,
        "brand": "apple",
        "price": "拾圆整"}]}

b = Box(box, force_format=True)
print(b.ipads[1].price)





import os
import json
import time
import glob
import email
import types
import inspect

from asyncio import Future
from functools import reduce, wraps
from collections.abc import Mapping
from argparse import Action, _SubParsersAction

from apistar import Include
from werkzeug._compat import string_types
from werkzeug.utils import escape, text_type
from werkzeug.http import dump_cookie, dump_header, parse_set_header

from .bases.session import Session
from .bases.components import Component


def get_real_method(obj, name):
    """
    IPython bug fix.
    """
    import types
    import inspect
    try:
        canary = getattr(obj, '_ipython_canary_method_should_not_exist_', None)
        if isinstance(canary, Future):
            canary.cancel()
    except Exception:
        return None

    if canary is not None:
        # It claimed to have an attribute it should never have
        return None

    try:
        m = getattr(obj, name, None)
        if isinstance(canary, Future):
            canary.cancel()
    except Exception:
        return None

    if inspect.isclass(obj) and not isinstance(m, types.MethodType):
        return None

    if callable(m):
        return m

    return None


def enhance_response(resp):

    def set_cookie(self, key, value='', max_age=None, expires=None,
                   path='/', domain=None, secure=False, httponly=False,
                   samesite=None):
        """Sets a cookie. The parameters are the same as in the cookie `Morsel`
        object in the Python standard library but it accepts unicode data, too.

        A warning is raised if the size of the cookie header exceeds
        :attr:`max_cookie_size`, but the header will still be set.

        :param key: the key (name) of the cookie to be set.
        :param value: the value of the cookie.
        :param max_age: should be a number of seconds, or `None` (default) if
                        the cookie should last only as long as the client's
                        browser session.
        :param expires: should be a `datetime` object or UNIX timestamp.
        :param path: limits the cookie to a given path, per default it will
                     span the whole domain.
        :param domain: if you want to set a cross-domain cookie.  For example,
                       ``domain=".example.com"`` will set a cookie that is
                       readable by the domain ``www.example.com``,
                       ``foo.example.com`` etc.  Otherwise, a cookie will only
                       be readable by the domain that set it.
        :param secure: If `True`, the cookie will only be available via HTTPS
        :param httponly: disallow JavaScript to access the cookie.  This is an
                         extension to the cookie standard and probably not
                         supported by all browsers.
        :param samesite: Limits the scope of the cookie such that it will only
                         be attached to requests if those requests are
                         "same-site".
        """
        self.headers['Set-Cookie'] = dump_cookie(
            key,
            value=value,
            max_age=max_age,
            expires=expires,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            charset=self.charset,
            max_size=self.max_cookie_size,
            samesite=samesite
        )

    def delete_cookie(self, key, path='/', domain=None):
        """Delete a cookie.  Fails silently if key doesn't exist.

        :param key: the key (name) of the cookie to be deleted.
        :param path: if the cookie that should be deleted was limited to a
                     path, the path has to be defined here.
        :param domain: if the cookie that should be deleted was limited to a
                       domain, that domain has to be defined here.
        """
        self.set_cookie(key, expires=0, max_age=0, path=path, domain=domain)

    setattr(resp, "set_cookie", set_cookie)
    setattr(resp, "delete_cookie", delete_cookie)
    setattr(resp, "max_cookie_size", 4093)

    def _set_property(name, doc=None):
        def fget(self):
            def on_update(header_set):
                if not header_set and name in self.headers:
                    del self.headers[name]
                elif header_set:
                    self.headers[name] = header_set.to_header()

            return parse_set_header(self.headers.get(name), on_update)

        def fset(self, value):
            if not value:
                del self.headers[name]
            elif isinstance(value, string_types):
                self.headers[name] = value
            else:
                self.headers[name] = dump_header(value)

        return property(fget, fset, doc=doc)

    setattr(resp, "vary", _set_property(
        "vary", doc='''
         The Vary field value indicates the set of request-header fields that
         fully determines, while the response is fresh, whether a cache is
         permitted to use the response to reply to a subsequent request
         without revalidation.'''))


def find_children(cls=Component, initialize=True):
    """
    获取所有(component)的子类或其实例。
    :param cls: 父类
    :param initialize: 是否生成实例
    :return:
    """
    def _load(cs):
        children = []
        for c in cs:
            children.append(c)
            children.extend(_load(c.__subclasses__()))
        return children

    loaded, unloaded = [], []
    for child in _load(cls.__subclasses__()):
        if initialize and getattr(child, "_instance", None):
            loaded.append(child._instance)
        else:
            unloaded.append(child)

    return [c() for c in unloaded] + loaded if initialize else unloaded


def logging_format(method, parttern, name, ca_name):
    return f"Route method: {method}, url: {parttern} to {name}."


def print_routing(routes, write=print, format=logging_format):
    for route, parents in walk_route(routes):
        name = reduce(lambda x, y: f"{x}:{y.name}", parents[1:], "view")\
               + ":" + route.name
        cont = route.controller.__class__
        ca_name = f"{cont.__module__}:{cont.__name__}#{route.handler.__name__}"
        pattern = reduce(
            lambda x, y: x.rstrip('/') + y.url, parents, "").rstrip("/")\
                  + route.url
        write(format(route.method, pattern, name, ca_name))


def walk_route(routes, parents=None):
    for route in routes:
        if parents is None:
            current_parents = []
        else:
            current_parents = parents[:]
        if isinstance(route, Include):
            current_parents.append(route)
            yield from walk_route(route.routes, current_parents)
        else:
            yield route, parents


def load_packages(current_path):
    """
    加载当前路径下的所有package，使得其中的Vontroller子类得以激活
    加载一个包时，如果包下面有子包，只需要导入子包，父包也会一起
    被加载。项目约定Controller子类必须定义在包中(__init__.py)。
    所以只考虑加载所有包，不考虑加载其它模块。
    :param current_path:
    :return:
    """
    files = glob.glob(os.path.join(current_path, "*"))
    find_dir = False

    for file in files:
        if os.path.isdir(file):
            find_dir = True
            if not load_packages(file):
                __import__(file.replace("/", ".").strip("."))

    return find_dir


def routing(controller):
    """
    获取当前Controller下所有route及其子Controller组成的Include
    :param controller:
    :return:
    """
    if not hasattr(controller, "prefix"):
        raise RuntimeError(f"{controller} is not routed! ")
    routes = []
    instance = controller()
    for name in vars(controller).keys():
        prop = getattr(controller, name)
        if hasattr(prop, "routes"):
            for route in prop.routes:
                route.controller = instance
            routes.extend(prop.routes)

    for child_controller in controller.__subclasses__():
        child_include = routing(child_controller)
        if child_include:
            routes.append(child_include)

    if routes:
        return Include(controller.prefix, controller.name, routes)


class ArgparseHelper(Action):
    """
        显示格式友好的帮助信息
    """

    def __init__(self,
                 option_strings,
                 dest="",
                 default="",
                 help=None):
        super(ArgparseHelper, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help()
        subparsers_actions = [
            action for action in parser._actions
            if isinstance(action, _SubParsersAction)]
        for subparsers_action in subparsers_actions:
            for choice, subparser in subparsers_action.choices.items():
                print("Command '{}'".format(choice))
                print(subparser.format_usage())

        parser.exit()


def parse_date(ims):
    """
    Parse rfc1123, rfc850 and asctime timestamps and return UTC epoch.
    """
    try:
        ts = email.utils.parsedate_tz(ims)
        return time.mktime(ts[:8] + (0,)) - (ts[9] or 0) - time.timezone
    except (TypeError, ValueError, IndexError, OverflowError):
        return None


def parse_range_header(header, maxlen=0):
    """
    Yield (start, end) ranges parsed from a HTTP Range header. Skip
        unsatisfiable ranges. The end index is non-inclusive.
    :param header:
    :param maxlen:
    :return:
    """
    if not header or header[:6] != 'bytes=': return
    ranges = [r.split('-', 1) for r in header[6:].split(',') if '-' in r]
    for start, end in ranges:
        try:
            if not start:  # bytes=-100    -> last 100 bytes
                start, end = max(0, maxlen-int(end)), maxlen
            elif not end:  # bytes=100-    -> all but the first 99 bytes
                start, end = int(start), maxlen
            else:          # bytes=100-200 -> bytes 100-200 (inclusive)
                start, end = int(start), min(int(end)+1, maxlen)
            if 0 <= start < end <= maxlen:
                yield start, end
        except ValueError:
            pass


async def file_iter_range(fp, offset, bytes, maxread=1024*1024):
    """
    Yield chunks from a range in a file. No chunk is bigger than maxread.
    :param fp:
    :param offset:
    :param bytes:
    :param maxread:
    :return:
    """
    await fp.seek(offset)
    while bytes > 0:
        part = await fp.read(min(bytes, maxread))
        if not part: break
        bytes -= len(part)
        yield part


class TypeEncoder(json.JSONEncoder):
    options = {Mapping: dict}

    @classmethod
    def register(cls, type_mapping):
        cls.options.update(type_mapping)

    def default(self, obj):
        for key, val in self.options.items():
            if isinstance(obj, key):
                return val(obj)
        return json.JSONEncoder.default(self, obj)


class HookReturn(Exception):
    pass


def redirect(location, code=302, Response=None):
    """Returns a response object (a ASGI application) that, if called,
    redirects the client to the target location.  Supported codes are 301,
    302, 303, 305, and 307.  300 is not supported because it's not a real
    redirect and 304 because it's the answer for a request with a request
    with defined If-Modified-Since headers.

    .. versionadded:: 0.6
       The location can now be a unicode string that is encoded using
       the :func:`iri_to_uri` function.

    .. versionadded:: 0.10
        The class used for the Response object can now be passed in.

    :param location: the location the response should redirect to.
    :param code: the redirect status code. defaults to 302.
    :param class Response: a Response class to use when instantiating a
        response. The default is :class:`werkzeug.wrappers.Response` if
        unspecified.
    """
    if Response is None:
        from apistar.http import HTMLResponse as Response

    display_location = escape(location)
    if isinstance(location, text_type):
        # Safe conversion is necessary here as we might redirect
        # to a broken URI scheme (for instance itms-services).
        from werkzeug.urls import iri_to_uri
        location = iri_to_uri(location, safe_conversion=True)
    response = Response(
        '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\n'
        '<title>Redirecting...</title>\n'
        '<h1>Redirecting...</h1>\n'
        '<p>You should be redirected automatically to target URL: '
        '<a href="%s">%s</a>.  If not click the link.' %
        (escape(location), display_location),
        headers={"Location": location},
        status_code=code)
    return response


def require(
        container_cls=Session,
        judge=lambda x: x.user,
        error="Login required!"):
    """
    装饰一个可被注入的函数，注入container_cls的实例，
    并调用judge判断其是否符合条件，否则抛出异常，
    异常信息为error。
    :param container_cls:
    :param judge:
    :param error:
    :return:
    """
    def auth(func):
        args = inspect.getfullargspec(func).args
        args_def = ", ".join(args)
        func_def = """
@wraps(func)
async def wrapper(__container, {}):
    from collections.abc import Awaitable
    assert judge(__container), (401, "{}")
    awaitable = func({})
    if isinstance(awaitable, Awaitable):
        return await awaitable
    return awaitable
        """.format(args_def, error, args_def)

        namespace = dict(__name__='entries_%s' % func.__name__)
        namespace["func"] = func
        namespace["wraps"] = wraps
        namespace["judge"] = judge
        exec(func_def, namespace)
        wrapper = namespace["wrapper"]
        new_func = types.FunctionType(
            wrapper.__code__,
            wrapper.__globals__,
            wrapper.__name__,
            func.__defaults__)
        new_func.__annotations__.update(func.__annotations__)
        new_func.__annotations__["__container"] = container_cls
        return new_func

    return auth



# -*- coding:utf-8 -*-
import os
import re
try:
    from setuptools import setup, find_packages
except:
    from distutils.core import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    mth = re.search("__version__\s?=\s?['\"]([^'\"]+)['\"]", init_py)
    if mth:
        return mth.group(1)
    else:
        raise RuntimeError("Cannot find version!")


def install_requires():
    """
    Return requires in requirements.txt
    :return:
    """
    try:
        with open("requirements.txt") as f:
            return [line.strip() for line in f.readlines() if line.strip()]
    except OSError:
        return []

try:
    LONG_DESCRIPTION = open("README.rst").read()
except UnicodeDecodeError:
    LONG_DESCRIPTION = open("README.rst", encoding="utf-8").read()

setup(
    name="star-builder",
    version=get_version("star_builder"),
    description="enhance apistar web framework. ",
    long_description=LONG_DESCRIPTION,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
    keywords="apistar",
    author="cn",
    author_email="cnaafhvk@foxmail.com",
    url="https://www.github.com/ShichaoMa/star_builder",
    entry_points={
        'console_scripts': [
            'apistar-create = star_builder:main',
            'apistar-console = star_builder:console',
            'apistar-routes = star_builder:show_routes'
        ],
    },
    license="MIT",
    packages=find_packages(),
    install_requires=install_requires(),
    include_package_data=True,
    zip_safe=True,
)



import requests


url = "http://127.0.0.1:8000/test_upload"

resp = requests.post(url, files={"test": open("/Users/mashichao/Downloads/zgtj.zip", "rb"),
                                 "test1": open("/Users/mashichao/Downloads/asyncio.png", "rb")})

print(resp.text)